import bpy
from bpy.types import Operator

import os
import sys
import subprocess
import pkg_resources

from pathlib import Path

from . import addon_id
from . import bl_info


addon_path = Path(__file__).parent
modules_dir = bpy.utils.user_resource("SCRIPTS", path="modules")
requirements_file = addon_path / "requirements.txt"


class DependencyPopup(Operator):
    bl_idname = f"{addon_id}.dependency_popup"
    bl_label = ""
    bl_description = "Install the dependencies required by this addon"
    bl_options = {"INTERNAL"}

    dep_str: bpy.props.StringProperty()
    feature: bpy.props.StringProperty()

    def invoke(self, context, event):
        self.deps = self.dep_str.split(",") if self.dep_str else requirements()
        if check(self.deps):
            self.report({"INFO"}, "Dependencies already installed!")
            return {"FINISHED"}

        return context.window_manager.invoke_props_dialog(self, width=400)

    def execute(self, context):
        if install(self.deps):
            self.report({"INFO"}, "Dependencies installed successfully!")
            return {"FINISHED"}

        else:
            self.report({"ERROR"}, "Could not install dependencies. Please check your internet connection!")
            return {"CANCELLED"}

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.alignment = "CENTER"
        row.label(text=f"{self.feature if self.feature else bl_info['name']} requires the following dependencies:")
        row = layout.row()
        row.alignment = "CENTER"
        box = row.box()
        for dep in self.deps:
            col = box.column()
            col.scale_y = 0.8
            col.label(text=dep, icon="DOT")

        layout.separator()
        row = layout.row()
        row.alignment = "CENTER"
        row.label(text="Do you want to install them now?")
        row = layout.row()
        row.alignment = "CENTER"
        row.scale_y = 0.3
        row.label(text="(This is a one time process, it will take a moment.)")
        layout.separator(factor=2)


def check(deps):
    if not Path(modules_dir).is_dir():
        return False

    try:
        for dep in deps:
            pkg_resources.require(dep)

        return True

    except Exception as e:
        print(f"Exception: {e}")

    return False


def install(deps):
    if check(deps):
        return True

    if not Path(modules_dir).is_dir():
        os.makedirs(modules_dir)

    cmd = [sys.executable, "-m", "pip", "install", *deps, "--target", os.fspath(modules_dir)]
    try:
        subprocess.check_call(cmd)

    except subprocess.SubprocessError as e:
        print(f"Exception: {e}")
        return False

    return True


def requirements():
    with requirements_file.open() as requirements:
        requirements = pkg_resources.parse_requirements(requirements)
        deps = [dep.project_name for dep in requirements]

    return deps
