import bpy
from bpy_extras import view3d_utils

from mathutils import Vector, Matrix


def deselect_all_bm(bm):
    for v in bm.verts:
        v.select = False
    for e in bm.edges:
        e.select = False
    for f in bm.faces:
        f.select = False


def deselect_all_mesh(mesh):
    for v in mesh.vertices:
        v.select = False
    for e in mesh.edges:
        e.select = False
    for f in mesh.polygons:
        f.select = False


def combine_cross_section_objs(context, sb, plane):
    col = bpy.data.collections.get("cross_sections")
    if not col:
        return
    objs = [o for o in col.objects.values() if o.name.startswith(f"{plane.name}.cross_section")]
    if not objs:
        return
    bpy.ops.object.select_all(action="DESELECT")

    new_objs = []
    for o in objs:
        new_obj = o.copy()
        new_obj.data = o.data.copy()
        context.scene.collection.objects.link(new_obj)
        new_obj.select_set(True)
        new_objs.append(new_obj)

    cs_obj = new_objs[0]
    context.view_layer.objects.active = cs_obj
    bpy.ops.object.join()
    return cs_obj


def is_modifier(event: bpy.types.Event, modifier):
    if modifier == "ALT" and event.alt:
        return True
    elif modifier == "SHIFT" and event.shift:
        return True
    elif modifier == "CTRL" and event.ctrl:
        return True
    elif modifier == "CMD" and event.oskey:
        return True
    else:
        return False


def get_bounds(context, objects, empty_sb_size):
    if not objects:
        i = empty_sb_size / 2
        v0 = Vector((-i, -i, -i))
        v1 = Vector((i, -i, -i))
        v2 = Vector((i, i, -i))
        v3 = Vector((-i, i, -i))

        v4 = Vector((-i, -i, i))
        v5 = Vector((i, -i, i))
        v6 = Vector((i, i, i))
        v7 = Vector((-i, i, i))

    else:
        # get vertex coos of all objects
        vert_cos = []
        active = context.view_layer.objects.active
        for o in objects:
            [setattr(m, "show_viewport", False) for m in o.modifiers if m.name.startswith("section_box_")]
            evaluated_obj = o.evaluated_get(context.evaluated_depsgraph_get())
            vert_cos.extend([evaluated_obj.matrix_world @ v.co for v in evaluated_obj.data.vertices])
            [setattr(m, "show_viewport", True) for m in o.modifiers if m.name.startswith("section_box_")]

        context.view_layer.objects.active = active

        offset = 0.0001

        # set up vertices
        max_x = max([vc[0] for vc in vert_cos]) + offset
        max_y = max([vc[1] for vc in vert_cos]) + offset
        max_z = max([vc[2] for vc in vert_cos]) + offset

        min_x = min([vc[0] for vc in vert_cos]) - offset
        min_y = min([vc[1] for vc in vert_cos]) - offset
        min_z = min([vc[2] for vc in vert_cos]) - offset

        v0 = Vector((min_x, min_y, min_z))
        v1 = Vector((max_x, min_y, min_z))
        v2 = Vector((max_x, max_y, min_z))
        v3 = Vector((min_x, max_y, min_z))

        v4 = Vector((min_x, min_y, max_z))
        v5 = Vector((max_x, min_y, max_z))
        v6 = Vector((max_x, max_y, max_z))
        v7 = Vector((min_x, max_y, max_z))

    return [v0, v1, v2, v3, v4, v5, v6, v7]


def get_rotation_matrix(v1: Vector, v2: Vector):
    v = v1.cross(v2)
    s = v.length  # sine of angle v1, v2
    c = v1.dot(v2)  # cosine of angle v1, v2

    skew_sym_m = Matrix(
        (
            (0, -v[2], v[1]),
            (v[2], 0, -v[0]),
            (-v[1], v[0], 0),
        )
    )

    rot_m = Matrix.Identity(3) + skew_sym_m + (skew_sym_m * skew_sym_m) * (1 / (1 + c))
    return rot_m


def get_sb_objs(context, sb, types):
    return [o for o in context.view_layer.objects if o.get(sb.name) and o.type in types and o.data.users != 0]


def mouse_to_plane3d(mouse_co, origin, normal):
    p1 = view3d_utils.region_2d_to_location_3d(
        bpy.context.region, bpy.context.space_data.region_3d, mouse_co, origin + normal
    )
    p2 = view3d_utils.region_2d_to_location_3d(
        bpy.context.region, bpy.context.space_data.region_3d, mouse_co, origin - normal
    )
    v = p2 - p1
    dotp = normal.dot(v)
    if dotp == 0:
        return False

    x = normal.dot(origin - p1) / dotp
    return p1 + (x * v)


def mouse_hit_3d(targets, mouse_co, radius):
    hits = {}
    for key, value in targets.items():
        t_2d = view3d_utils.location_3d_to_region_2d(bpy.context.region, bpy.context.space_data.region_3d, value)
        dst = (t_2d - mouse_co).length
        if dst < radius:
            hits[key] = dst

    if not hits:
        return None

    return min(hits, key=hits.get)


def mouse_hit_2d(targets, mouse_co, radius):
    hits = {}
    for i, t in enumerate(targets):
        if t is None:
            continue
        dst = (t - mouse_co).length
        if dst < radius:
            hits[i] = dst

    if not hits:
        return None

    return min(hits, key=hits.get)


def line_line_intersection_2d(a, b, c, d):
    if abs((b - a).normalized().dot((d - c).normalized())) == 1:
        return False

    ab = b - a
    cd = d - c

    s = (c - a).cross(cd) / ab.cross(cd)
    t = (a - c).cross(ab) / cd.cross(ab)

    if 0 < s < 1 and 0 < t < 1:
        return True

    return False


def tri_area_2d(v1, v2, v3):
    mat = Matrix(((v1[0], v1[1], 1), (v2[0], v2[1], 1), (v3[0], v3[1], 1)))
    return abs(mat.determinant() / 2)


def point_tri_intersection_2d(p, v1, v2, v3):
    a1 = tri_area_2d(p, v1, v2)
    a2 = tri_area_2d(p, v1, v3)
    a3 = tri_area_2d(p, v2, v3)

    area = a1 + a2 + a3
    return abs(area - tri_area_2d(v1, v2, v3)) < 0.1
