import bpy
import gpu
import numpy as np

from bpy_extras import view3d_utils
from mathutils import Vector, Matrix
from gpu_extras.batch import batch_for_shader
from pathlib import Path

from .load_svg import load_svg
from ..utils.sectionbox import get_sectionbox
from ..utils.preferences import get_prefs


class DrawHandles:
    def __init__(self, context):
        self.prefs = get_prefs()
        self.init_shader()
        self.handler = bpy.types.SpaceView3D.draw_handler_add(
            self.draw_callback, (context,), "WINDOW", "POST_PIXEL"
        )

    def init_shader(self):
        self.shader = gpu.shader.from_builtin("UNIFORM_COLOR")
        self.shader.bind()

        path = Path(__file__).parent.parent.absolute().joinpath("icons", "handle.svg")
        self.handle_icon = load_svg(path)

    def draw_callback(self, context):
        sb = get_sectionbox(context, check_active=True)
        if not sb:
            return

        gpu.state.blend_set("ALPHA")
        euler = context.region_data.view_rotation.to_euler()
        rot = context.region_data.view_rotation @ Vector((0, 0, -1))
        rx = -euler[2] if rot[2] < 0 else euler[2]
        ry = -euler[2] + np.pi / 2 if rot[2] < 0 else euler[2] + np.pi / 2
        rots = (np.pi / 2, np.pi / 2, ry, ry, rx, rx)
        if not context.window_manager.sectionbox.edge_select:
            for plane in sb.planes:
                face = sb.object.data.polygons[plane.id]
                pos = view3d_utils.location_3d_to_region_2d(
                    context.region,
                    context.space_data.region_3d,
                    sb.object.matrix_world @ face.center,
                )
                scale = (
                    self.prefs.handle_hover_scale
                    if plane.id == context.window_manager.sectionbox.hovering_plane
                    else 1
                )
                scale_mat = Matrix.Scale(scale, 2)
                rot_mat = Matrix.Rotation(rots[plane.id], 2)
                cos = [pos + scale_mat @ rot_mat @ co for co in self.handle_icon[0]]

                try:
                    color = self.prefs.handle_color
                except Exception as e:
                    print(e)

                if plane.id == sb.active_plane:
                    color = self.prefs.handle_color_active

                self.shader.uniform_float("color", color)

                batch = batch_for_shader(
                    self.shader, "TRIS", {"pos": cos}, indices=self.handle_icon[1]
                )
                batch.draw(self.shader)

        else:
            plane = sb.planes[sb.active_plane]
            face = sb.object.data.polygons[plane.id]
            for i, edge in enumerate(
                [
                    sb.object.data.edges[sb.object.data.edge_keys.index(ek)]
                    for ek in face.edge_keys
                ]
            ):
                edge_center = (
                    sb.object.data.vertices[edge.vertices[0]].co
                    + sb.object.data.vertices[edge.vertices[1]].co
                ) / 2
                pos = view3d_utils.location_3d_to_region_2d(
                    context.region,
                    context.space_data.region_3d,
                    sb.object.matrix_world @ edge_center,
                )

                scale = (
                    self.prefs.handle_hover_scale
                    if i == context.window_manager.sectionbox.hovering_edge
                    else 1
                )
                scale_mat = Matrix.Scale(scale, 2)
                rot_mat = Matrix.Rotation(rots[plane.id], 2)
                cos = [pos + scale_mat @ rot_mat @ co for co in self.handle_icon[0]]
                self.shader.uniform_float("color", (0, 0, 0, 1))
                batch = batch_for_shader(
                    self.shader, "TRIS", {"pos": cos}, indices=self.handle_icon[1]
                )
                batch.draw(self.shader)

    def remove_handler(self):
        bpy.types.SpaceView3D.draw_handler_remove(self.handler, "WINDOW")
